
#include <vector>
#include <algorithm>
#include <iterator>
#include <numeric>
#include <set>
#include <iostream>
//#include "design.h"
#include <string.h>

extern "C" {
#include "mainInt.h"
#include "abc_decl.h"
#include "extra.h"
}


#define _SOURCE_SINK_EDGE_
#define _INC_SIM_	// incremental simulation
#define __int64 long long


class FaultSim {

	public: 

		Abc_Ntk_t* pNtk;

		int nVectors;
		int nSimIters;

		std::vector<unsigned*> vTruth;			// truth table for all internal nodes
		std::vector<unsigned __int64> vSim;		// temporary sim vector for all nodes
		std::vector<std::vector<unsigned __int64> > vRandCi;	
		// random vectors generated in CIs
		std::vector<std::vector<unsigned __int64> > vGoldenCo;
		// simulation results for COs in the golden case
		std::vector<std::vector<unsigned __int64> > vGoldenSim;
		// simulation results for internal nodes the golden case
		Vec_Ptr_t* vSortedNodes;				// topologically sorted internal nodes
		std::vector<float> vBitCrit;			// criticality of configuration bit
		std::vector<std::vector<float> > vECrit;// criticality of an edge (from one LUT to the other)

		std::vector<Abc_Obj_t*> vFaultyNode_TFO;// a temp vector to store transitive fanout cone of the current faulty node

		bool* vFaultyNode_COs;					// a temp vector to store all COs covered by the current faulty node

		std::vector<bool> vOnOffState;			// vOnOffState[i] = 0 iff LUT i's truth table is ON dominate
		std::vector<std::vector<std::vector<bool> > > vOnOffSet;
		// vOnOffSet[i][j][k] = 0 iff LUT i's ON/OFF set's jth entrance's kth fanin is 0
		char** vIdMap; //for mapping ID in ABC to the real ones. 
		int   vNumId;
	public: 

		FaultSim(Abc_Ntk_t* pNtk2Sim, int nVec) 
		{
			nVectors = nVec;
			nSimIters = nVectors / 64;

			pNtk = pNtk2Sim;

			vTruth.resize(Abc_NtkObjNum(pNtk));
			vSim.resize(Abc_NtkObjNum(pNtk));
			vIdMap = (char**)malloc((Abc_NtkObjNum(pNtk)*sizeof(char*)));
			vRandCi.resize(nSimIters);
			for(std::vector<std::vector<unsigned __int64> >::iterator vvIter = vRandCi.begin();
					vvIter != vRandCi.end(); ++ vvIter) {
				vvIter->resize(Abc_NtkCiNum(pNtk));
			}

			vGoldenCo.resize(nSimIters);
			for(std::vector<std::vector<unsigned __int64> >::iterator vvIter = vGoldenCo.begin();
					vvIter != vGoldenCo.end(); ++ vvIter) {
				vvIter->resize(Abc_NtkCoNum(pNtk));
			}

			vGoldenSim.resize(nSimIters);
			for(std::vector<std::vector<unsigned __int64> >::iterator vvIter = vGoldenSim.begin();
					vvIter != vGoldenSim.end(); ++ vvIter) {
				vvIter->resize(Abc_NtkObjNum(pNtk));
			}

			vFaultyNode_TFO.reserve(Abc_NtkObjNum(pNtk));
			vFaultyNode_COs = new bool[Abc_NtkObjNum(pNtk)];

			vOnOffState.assign(Abc_NtkObjNum(pNtk), false);
			vOnOffSet.resize(Abc_NtkObjNum(pNtk));

		}

		~FaultSim()
		{
			for(std::vector<unsigned*>::iterator uIter = vTruth.begin();
					uIter != vTruth.end(); ++ uIter) {
				if(*uIter) delete [] *uIter;
			}

			Vec_PtrFree(vSortedNodes);

			delete [] vFaultyNode_COs;
		}

		inline unsigned __int64 my_rand_u64()
		{
			// rand() generates 0~RAND_MAX(0x7fff)

			return	((unsigned __int64 ) rand()) << 48 ^
				((unsigned __int64 ) rand()) << 32 ^ 
				((unsigned __int64 ) rand()) << 16 ^
				(unsigned __int64 ) rand();
		}

		inline void my_printHex_u64(unsigned __int64 num)
		{
			Extra_PrintHex(stdout, num >> 32, 5);
			Extra_PrintHex(stdout, num, 5);
			printf("\n");
		}

#ifdef _YH_DEBUG_
		void print_vector_nodes()
		{
			Abc_Obj_t* pNode, *pFanin;
			void* pObj;
			int i, f;
			std::set<Abc_Obj_t*> sCOs;

			Vec_PtrForEachEntry(vSortedNodes, pObj, i)
			{
				pNode = (Abc_Obj_t*) pObj;

				Extra_PrintHex(stdout, vTruth[pNode->Id][0], Abc_ObjFaninNum(pNode));

				printf(", %d (%d): ", pNode->Id, pNode->Type);
				Abc_ObjForEachFanin(pNode, pFanin, f)
				{
					printf(" %d", pFanin->Id);
				}
				printf("|");
				Abc_ObjForEachFanout(pNode, pFanin, f)
				{
					if(Abc_ObjIsCo(pFanin)) sCOs.insert(pFanin);
					printf(" %d", pFanin->Id);
				}
				printf("\n");
			}
			printf("==================\n");

			std::set<Abc_Obj_t*> checkedNodes;

			Vec_PtrForEachEntry(vSortedNodes, pObj, i)
			{
				pNode = (Abc_Obj_t*) pObj;
				// printf("%d\n", pNode->Id);
				checkedNodes.insert(pNode);

				Abc_ObjForEachFanin(pNode, pFanin, f)
				{
					if(checkedNodes.find(pFanin) == checkedNodes.end()) {
						printf("%d (%d)-> %d (%d)\n", pFanin->Id, pFanin->Type, pNode->Id, pNode->Type);
					}
				}
			}
			printf("CO=====================\n");

			Abc_NtkForEachCo(pNtk, pNode, i)
			{
				if(sCOs.find(pNode) == sCOs.end()) {
					printf("%d (%d)-> %d (%d)\n", 
							Abc_ObjFanin0(pNode)->Id, Abc_ObjFanin0(pNode)->Type, 
							pNode->Id, pNode->Type);
				}
			}
		}
#endif

		inline void simulate64_oneNode(Abc_Obj_t* pNode, Abc_Obj_t*  pFaultyNode)
		{
			unsigned __int64 tmp_u64, minTerm_u64;
			int nFanins = Abc_ObjFaninNum(pNode);
			unsigned* pTruth = vTruth[pNode->Id];

			// consider simple cases
			if (Extra_TruthIsConst0(pTruth, nFanins) )
			{
				vSim[pNode->Id] = 0;
				return;
			}
			if (Extra_TruthIsConst1(pTruth, nFanins) )
			{
				vSim[pNode->Id] = 0xffffffffffffffffULL;
				return;
			}
			if (nFanins == 1 && pFaultyNode != pNode) 
			{
				vSim[pNode->Id] = Abc_NodeIsBuf(pNode) ? 
					vSim[Abc_ObjFanin0(pNode)->Id] : 
					~vSim[Abc_ObjFanin0(pNode)->Id] ;
				return;
			}

			tmp_u64 = 0;

#ifdef _YH_TRUTH_PRE_
			for(std::vector<std::vector<bool> >::iterator vvIter = vOnOffSet[pNode->Id].begin();
					vvIter != vOnOffSet[pNode->Id].end(); ++ vvIter) {

				minTerm_u64 = 0xffffffffffffffffULL;

				for(int fanin = 0; fanin < nFanins; fanin ++ ) {
					if((*vvIter)[fanin])
						minTerm_u64 &= vSim[Abc_ObjFanin(pNode, fanin)->Id];
					else
						minTerm_u64 &= ~vSim[Abc_ObjFanin(pNode, fanin)->Id];
				}
				tmp_u64 |= minTerm_u64;
			}
			if(!vOnOffState[pNode->Id]) tmp_u64 = ~tmp_u64;
#else
			for(int truthEntry = 0; truthEntry < (1<<nFanins); truthEntry ++ ) {
				if(!Extra_TruthHasBit(pTruth, truthEntry)) continue;
				minTerm_u64 = 0xffffffffffffffffULL;
				for(int fanin = 0; fanin < nFanins; fanin ++ ) {
					if(truthEntry & (1 << fanin))
						minTerm_u64 &= vSim[Abc_ObjFanin(pNode, fanin)->Id];
					else
						minTerm_u64 &= ~vSim[Abc_ObjFanin(pNode, fanin)->Id];
				}
				tmp_u64 |= minTerm_u64;
			}
#endif // _YH_TRUTH_PRE_
			vSim[pNode->Id] = tmp_u64;
#ifdef _YH_DEBUG_
			printf("vSim[%d]: ", pNode->Id);
			my_printHex_u64(vSim[pNode->Id]);
#endif
		}

		inline void simulate64_processTruthTable()
		{
			unsigned __int64 tmp_u64, minTerm_u64;
			void* pObj;
			int i;
			Abc_Obj_t* pNode;

			// go through all internal nodes
			Vec_PtrForEachEntry(vSortedNodes, pObj, i)
			{
				pNode = (Abc_Obj_t*) pObj;

				unsigned* pTruth = vTruth[pNode->Id];
				int nFanins = Abc_ObjFaninNum(pNode);

				// count number of ones
				// TODO: there are better ways to do so, e.g., using Extra_TruthCountOnes
				// however, such way needs to truncate the front part of the truth table
				// for example, for nFanins == 4, the truth table could be 0xabcdabcd, 
				// since the minimal unit is unsigned (32bits)
				int numOnes = 0;
				for(int truthEntry = 0; truthEntry < (1<<nFanins); truthEntry ++ ) {
					if(Extra_TruthHasBit(pTruth, truthEntry)) numOnes ++;
				}
				if(numOnes < (1 << nFanins >> 1)) {
					vOnOffState[pNode->Id] = true;	// 0 dominate
					vOnOffSet[pNode->Id].resize(numOnes);
				}
				else {
					vOnOffState[pNode->Id] = false;	// 1 dominate
					numOnes = (1 << nFanins) - numOnes;
					vOnOffSet[pNode->Id].resize(numOnes);
				}

				for(int bitIter = 0; bitIter < numOnes; bitIter ++)
				{
					vOnOffSet[pNode->Id][bitIter].resize(nFanins);
				}

				int onOffCounter = 0;
				for(int truthEntry = 0; truthEntry < (1<<nFanins); truthEntry ++ ) {

					if(Extra_TruthHasBit(pTruth, truthEntry) != vOnOffState[pNode->Id]) continue;

					for(int fanin = 0; fanin < nFanins; fanin ++ ) {
						if(truthEntry & (1 << fanin))
							vOnOffSet[pNode->Id][onOffCounter][fanin] = true;
						else
							vOnOffSet[pNode->Id][onOffCounter][fanin] = false;
					}

					onOffCounter ++ ;
				} 
			}
		}

		inline void find_TFO_rec(Abc_Obj_t* pNode, std::vector<Abc_Obj_t*>& vTFO)
		{
			Abc_Obj_t* pFanout;
			int i;

			// if the node is already visited, skip
			if(Abc_NodeIsTravIdCurrent(pNode)) return;
			// mark the node as visited
			Abc_NodeSetTravIdCurrent(pNode);
			// skip the CO
			if(Abc_ObjIsCo(pNode)) {
				vFaultyNode_COs[pNode->Id] = 1;
				return;
			}
			assert(Abc_ObjIsNode(pNode));

			vTFO.push_back(pNode);

			// visit the transitive fanout of the node
			Abc_ObjForEachFanout(pNode, pFanout, i)
			{
				find_TFO_rec(pFanout, vTFO);
			}
		}

		inline void find_TFO(Abc_Obj_t* pNode, std::vector<Abc_Obj_t*>& vTFO)
		{
			// find transitive fanout cone of pNode

			Abc_Obj_t* pFanout;
			int i;

			assert(pNtk->ntkType == ABC_NTK_LOGIC);

			// reset the traversal ID
			Abc_NtkIncrementTravId(pNtk);
			Abc_NodeSetTravIdCurrent(pNode);
			Abc_ObjForEachFanout(pNode, pFanout, i)
			{
				find_TFO_rec(pFanout, vTFO);
			}
		}

		inline void simulate64_inc_TFO(Abc_Obj_t* pFaultyNode)
		{
			vFaultyNode_TFO.clear();
			vFaultyNode_TFO.push_back(pFaultyNode);

			memset(vFaultyNode_COs, 0, sizeof(bool) * Abc_NtkObjNum(pNtk));

			// find the TFO cone of pFaultyNode
			find_TFO(pFaultyNode, vFaultyNode_TFO);

#ifdef _YH_DEBUG_
			printf("%d: ", pFaultyNode->Id);
			for(std::vector<Abc_Obj_t*>::iterator vIter = vFaultyNode_TFO.begin(); vIter != vFaultyNode_TFO.end(); ++ vIter){
				Abc_Obj_t* pNode = *vIter;
				printf(" %d", pNode->Id);
			}
			printf("\n");
#endif
		}

		inline void find_covered_CO_rec(Abc_Obj_t* pNode)
		{
			Abc_Obj_t* pFanout;
			int i;

			// if the node is already visited, skip
			if(Abc_NodeIsTravIdCurrent(pNode)) return;
			// mark the node as visited
			Abc_NodeSetTravIdCurrent(pNode);

			if(Abc_ObjIsCo(pNode)) {
				vFaultyNode_COs[pNode->Id] = 1;
				return;
			}
			assert(Abc_ObjIsNode(pNode));

			// visit the transitive fanout of the node
			Abc_ObjForEachFanout(pNode, pFanout, i)
			{
				find_covered_CO_rec(pFanout);
			}
		}

		inline void simulate64_inc_CO(Abc_Obj_t* pFaultyNode)
		{
			// find COs covered by pNode

			memset(vFaultyNode_COs, 0, sizeof(bool) * Abc_NtkObjNum(pNtk));

			Abc_Obj_t* pFanout;
			int i;

			assert(pNtk->ntkType == ABC_NTK_LOGIC);

			// reset the traversal ID
			Abc_NtkIncrementTravId(pNtk);
			Abc_NodeSetTravIdCurrent(pFaultyNode);
			Abc_ObjForEachFanout(pFaultyNode, pFanout, i)
			{
				find_covered_CO_rec(pFanout);
			}
		}

		inline void simulate64_inc_cpy_vSim(int curSimIter)
		{
			int i;
			void* pObj;
			Abc_Obj_t* pNode;

			//int count = 0;
			// copy vGoldenSim to all nodes in {NODE / TFO}
			Vec_PtrForEachEntry(vSortedNodes, pObj, i)
			{
				pNode = (Abc_Obj_t*) pObj;

				// don't copy vSim if the node is in the TFO of pFaultyNode
				if(!Abc_NodeIsTravIdCurrent(pNode)) 
					vSim[pNode->Id] = vGoldenSim[curSimIter][pNode->Id];
				//else count ++ ;;
			}

			//assert(count == vFaultyNode_TFO.size()) ;

		}

		inline void simulate64_inc_do(Abc_Obj_t* pFaultyNode)
		{
			// Incrementally simulate the circuit, i.e.,
			// only fanout cone of pFaultyNode is simulated

			Abc_Obj_t* pNode;
			void* pObj;
			int i;

			/*int counter1 = 0, counter2 = 0;
			  std::set<int> set1, set2;*/

			/*for(std::vector<Abc_Obj_t*>::iterator ppNode = vFaultyNode_TFO.begin();
			  ppNode != vFaultyNode_TFO.end(); ++ ppNode) {

			  simulate64_oneNode(*ppNode, pFaultyNode);
			  counter1++;
			  set1.insert((*ppNode)->Id);
			  std::cout << " " << (*ppNode)->Id;
			  }
			  std::cout << std::endl;*/
			Vec_PtrForEachEntry(vSortedNodes, pObj, i)
			{
				pNode = (Abc_Obj_t*) pObj;

				// only simulate if the node is in the TFO of pFaultyNode
				if(Abc_NodeIsTravIdCurrent(pNode)) {
					simulate64_oneNode(pNode, pFaultyNode);
					/*counter2 ++;
					  set2.insert(pNode->Id);
					  std::cout << " " << pNode->Id;*/
				}
				//else count ++ ;;
			}
			/*std::cout << std::endl;
			  if(counter1 != counter2) {printf("%d, %d\n", counter1, counter2); exit(-1);}
			  std::set_difference(set1.begin(), set1.end(), set2.begin(), set2.end(),
			  std::ostream_iterator<int>(std::cout, " "));*/

		}

		inline void simulate64_inc_do_edge(Abc_Obj_t* pFaultyNode, Abc_Obj_t* pFaultyNodeFanout)
		{
			// Incrementally simulate the circuit, i.e.,
			// only fanout cone of pFaultyNode is simulated

			Abc_Obj_t* pNode;
			void* pObj;
			int i;

			//for(std::vector<Abc_Obj_t*>::iterator ppNode = vFaultyNode_TFO.begin();
			//	ppNode != vFaultyNode_TFO.end(); ++ ppNode) {

			//		// fault injection in an edge 
			//		if(pFaultyNodeFanout->Id == (*ppNode)->Id)
			//			vSim[pFaultyNode->Id] = ~vSim[pFaultyNode->Id];

			//		simulate64_oneNode(*ppNode, pFaultyNode);

			//		// fault correction in an edge 
			//		if(pFaultyNodeFanout->Id == (*ppNode)->Id)
			//			vSim[pFaultyNode->Id] = ~vSim[pFaultyNode->Id];

			//}
			Vec_PtrForEachEntry(vSortedNodes, pObj, i)
			{
				pNode = (Abc_Obj_t*) pObj;

				// only simulate if the node is in the TFO of pFaultyNode
				if(!Abc_NodeIsTravIdCurrent(pNode)) continue;

				// fault injection in an edge 
				if(pFaultyNodeFanout == pNode)
					vSim[pFaultyNode->Id] = ~vSim[pFaultyNode->Id];

				simulate64_oneNode(pNode, pFaultyNode);

				// fault correction in an edge 
				if(pFaultyNodeFanout == pNode)
					vSim[pFaultyNode->Id] = ~vSim[pFaultyNode->Id];
			}

		}

		inline void simulate64(Abc_Obj_t* pFaultyNode)
		{
			// Simulate the entire circuit with a 64-bit random vector

			void* pObj;
			int i;

			// propagate through all internal nodes
			Vec_PtrForEachEntry(vSortedNodes, pObj, i)
			{
				simulate64_oneNode((Abc_Obj_t*) pObj, pFaultyNode);
			}

		}

		float simulate_estimate_runtime(float runtime_oneIter)
		{
			float runtime = 0.0f;
			int iFaultyNode;
			void* pObj;
			Abc_Obj_t* pFaultyNode;

			Vec_PtrForEachEntry(vSortedNodes, pObj, iFaultyNode)
			{
				pFaultyNode = (Abc_Obj_t*) pObj;

				// get TFO for pFaultyNode
				// need to be done once for each LUT
				simulate64_inc_TFO(pFaultyNode);

				runtime += (float) (1 << Abc_ObjFaninNum(pFaultyNode)) * vFaultyNode_TFO.size() / vSortedNodes->nSize;
			}

			return runtime * runtime_oneIter * nSimIters ;
		}

		int simulate_all()
		{
			assert(pNtk->ntkFunc == ABC_FUNC_AIG);
			assert(pNtk->ntkType == ABC_NTK_LOGIC);

			// compute the truth table for all internal nodes

			Vec_Int_t* vTruthMem = Vec_IntAlloc(Abc_NtkObjNum(pNtk) * 2);
			unsigned* pTruth;
			int nConfigBits = 0, i, nFanins;
			Abc_Obj_t* pNode;

			vECrit.resize(Abc_NtkObjNum(pNtk));

			Abc_NtkForEachNode(pNtk, pNode, i)
			{
				assert(Abc_ObjIsNode(pNode));
				nFanins = Abc_ObjFaninNum(pNode);
				assert(nFanins <= 8);

				pTruth = Abc_ConvertAigToTruth(
						(Hop_Man_t *) pNode->pNtk->pManFunc, 
						Hop_Regular((Hop_Obj_t *) pNode->pData), 
						nFanins, vTruthMem, 0 );
				if(Hop_IsComplement((Hop_Obj_t *) pNode->pData) )
					Extra_TruthNot( pTruth, pTruth, nFanins );

				int nTruthWords;
				if(nFanins < 6) nTruthWords = 1;
				else nTruthWords = (1<< nFanins) >> 5;

				vTruth[pNode->Id] = new unsigned[nTruthWords];
				Extra_TruthCopy(vTruth[pNode->Id], pTruth, nFanins);
				nConfigBits += (1 << nFanins);

				vECrit[pNode->Id].assign(Abc_ObjFanoutNum(pNode), 0.0f);
			}
			vBitCrit.assign(nConfigBits, 0.0f);

			// prepare for simulation

			// topological sort of all internal nodes
			vSortedNodes = Abc_NtkDfs(pNtk, 0);

#ifdef _YH_TRUTH_PRE_
			// precompute the on/off set of the truth table for fast access
			simulate64_processTruthTable();
#endif // _YH_TRUTH_PRE_

#ifdef _YH_DEBUG_
			print_vector_nodes();
#endif

			// double check the topological sort results
			// 1. CO nodes are not included in the internal nodes
			// 2. CI nodes are not included in the internal nodes

			clock_t clk = clock();

			// simulate the golden case

			for(int nSim = 0; nSim < nSimIters; nSim ++) {

				// set initial random vectors for PIs
				Abc_NtkForEachCi(pNtk, pNode, i)
				{
					if(pNode->Type == ABC_OBJ_CONST1)
						vRandCi[nSim][i] = vSim[pNode->Id] = 0xffffffffffffffffULL;
					else {
						assert(pNode->Type == ABC_OBJ_PIO || pNode->Type == ABC_OBJ_PI || pNode->Type == ABC_OBJ_BO);
						vRandCi[nSim][i] = vSim[pNode->Id] = my_rand_u64();
					}
#ifdef _YH_DEBUG_
					printf("vRandCi[%d]: ", pNode->Id);
					my_printHex_u64(vRandCi[nSim][i]);
#endif
				}

				simulate64(NULL);

				// record the simulation results for all COs in the golden case
				Abc_NtkForEachCo(pNtk, pNode, i)
				{
					vGoldenCo[nSim][i] = vSim[Abc_ObjFanin0(pNode)->Id];
#ifdef _YH_DEBUG_
					printf("VGoldenCo[%d]: ", pNode->Id);
					my_printHex_u64(vGoldenCo[nSim][i]);
#endif
				}

				// record the simulation results for all internal nodes in the golden case
				Abc_NtkForEachNode(pNtk, pNode, i)
				{
					vGoldenSim[nSim][pNode->Id] = vSim[pNode->Id];
				}

			}
			/*float clk_end = clock();
			  printf("Estimated simulation time: %6.2f sec\n", 
			  simulate_estimate_runtime((float)(clk_end - clk) /(CLOCKS_PER_SEC)) );*/

			// simulate all faults

			Abc_Obj_t*pFaultyNode;
			void* pObj;
			int iFaultyNode, curFaultyBit;
			unsigned __int64 bitCheck, pre_bitCheck;

			int tfo_time = 0;
			int sim_time = 0;
			int cpy_time = 0;
			int inc_time = 0;
			int count_time = 0;

			clk = clock();
			for(int nSim = 0; nSim < nSimIters; nSim ++) 
			{
				// set initial random vectors for PIs
				// need to be done once for each nSim
				Abc_NtkForEachCi(pNtk, pNode, i)
				{
					vSim[pNode->Id] = vRandCi[nSim][i];
				}

				curFaultyBit = 0;

				Vec_PtrForEachEntry(vSortedNodes, pObj, iFaultyNode)
				{
					pFaultyNode = (Abc_Obj_t*) pObj;

					/*printf("%d: ", pFaultyNode->Id);
					  Abc_Obj_t* pFanout;
					  Abc_ObjForEachFanout(pFaultyNode, pFanout, i) {
					  printf(" %d", pFanout->Id);
					  }
					  printf("\n");*/

					// get TFO and COs covered by pFaultyNode
					// need to be done once for each LUT
					tfo_time -= clock();
					simulate64_inc_TFO(pFaultyNode);
					tfo_time += clock();

					// copy golden vSim to non-changed signals (nodes not in TFO of pFaultyNode),
					// need to be done once for each LUT and each nSim
					cpy_time -= clock();
					simulate64_inc_cpy_vSim(nSim); 
					cpy_time += clock();

					// get COs covered by pFaultyNode
					// simulate64_inc_CO(pFaultyNode);

					/* // pre-count the faulty vectors for unchanged COs
					   pre_bitCheck = 0;
					   Abc_NtkForEachCo(pNtk, pNode, i)
					   {
					   if(!vFaultyNode_COs[pNode->Id])
					   pre_bitCheck |= (vGoldenCo[nSim][i] ^ vSim[Abc_ObjFanin0(pNode)->Id]);
					   }
					   assert(!pre_bitCheck) ;*/

					sim_time -= clock();
					// simulate LUT faults
					for(int fb = 0; fb < (1 << Abc_ObjFaninNum(pFaultyNode)); ++ fb) {

						// fault injection
						Extra_TruthXorBit(vTruth[pFaultyNode->Id], fb);

						inc_time -= clock();
#ifdef _INC_SIM_
						simulate64_inc_do(pFaultyNode);
#else
						simulate64(pFaultyNode);
#endif
						inc_time += clock();

						// count the faulty vectors
						count_time -= clock();
						bitCheck = 0;
						Abc_NtkForEachCo(pNtk, pNode, i)
						{
#ifdef _INC_SIM_
							// only count the faulty vectors for changed COs
							if(vFaultyNode_COs[pNode->Id])
#endif
								bitCheck |= (vGoldenCo[nSim][i] ^ vSim[Abc_ObjFanin0(pNode)->Id]);
						}
						vBitCrit[curFaultyBit ++] += (float) 
							(Extra_WordCountOnes((unsigned) (bitCheck >> 32)) + 
							 Extra_WordCountOnes((unsigned) bitCheck )) / 64.0 ;
						count_time += clock();
#ifdef _YH_DEBUG_				
						printf("%d %d: %f\n", pFaultyNode->Id, fb, nFaultyVecs / 64.0 / nSimIters);
#endif
						// fault correction
						Extra_TruthXorBit(vTruth[pFaultyNode->Id], fb);
					}

#ifdef _SOURCE_SINK_EDGE_
					// simulate interconnect faults (i.e., fanouts of an LUT)
					for(int fb = 0; fb < Abc_ObjFanoutNum(pFaultyNode); ++ fb) {

						// simulate64(pFaultyNode);
						inc_time -= clock();
						simulate64_inc_do_edge(pFaultyNode, Abc_ObjFanout(pFaultyNode, fb));
						inc_time += clock();

						// count the faulty vectors
						count_time -= clock();
						bitCheck = 0;
						Abc_NtkForEachCo(pNtk, pNode, i)
						{
							// only count the faulty vectors for changed COs
							if(vFaultyNode_COs[pNode->Id])
								bitCheck |= (vGoldenCo[nSim][i] ^ vSim[Abc_ObjFanin0(pNode)->Id]);
						}
						vECrit[pFaultyNode->Id][fb] += (float) 
							(Extra_WordCountOnes((unsigned) (bitCheck >> 32)) + 
							 Extra_WordCountOnes((unsigned) bitCheck )) / 64.0 ; 
						count_time += clock();
					}
#endif // _SOURCE_SINK_EDGE_ 
					sim_time += clock();
				}
			}
			printf("tfo_time = %6.2f sec, sim_time = %6.2f sec, cpy_time = %6.2f, inc_time = %6.2f, count_time = %6.2f\n", 
					(float)(tfo_time)/(float)(CLOCKS_PER_SEC), (float)(sim_time)/(float)(CLOCKS_PER_SEC),
					(float)(cpy_time)/(float)(CLOCKS_PER_SEC), (float)(inc_time)/(float)(CLOCKS_PER_SEC),
					(float)(count_time)/(float)(CLOCKS_PER_SEC));

			printf("Real simulation time: %6.2f sec\n", 
					(float)(clock() - clk)/(float)(CLOCKS_PER_SEC));
			for(std::vector<float>::iterator fIter = vBitCrit.begin(); fIter != vBitCrit.end(); ++ fIter)
				*fIter /= nSimIters;

			float fSumECrit = 0.0f;
			int nEdges = 0;
			Abc_Obj_t *lnode;
			int lNumWires = 0;
			FILE *fp1 ;
			char lCmd [1024];
			sprintf(lCmd, "%s.sens",pNtk->pName);
			fp1 = fopen(lCmd, "w+");
			if(!fp1)
			{
				printf(" Cannot open the file <%s> in this directory\n",lCmd);
				exit(0);
			}


			Abc_NtkForEachNode(pNtk, pNode, i)
			{
				int iFanOut, iFanIn;
				//for(int iFanout = 0; iFanout < Abc_ObjFanoutNum(pNode); iFanout ++, nEdges ++)
				Abc_ObjForEachFanout(pNode, lnode, iFanOut)
				{

					//	printf(" The Criticality of wire %s to %s is %6.4f\n",
					//	vIdMap[pNode->Id], vIdMap[lnode->Id], vECrit[i][iFanOut]);
					//if(0)
					if((!strcmp(vIdMap[pNode->Id],vIdMap[lnode->Id])))
					{
						if(vECrit[i][iFanOut])
						{
							printf(" Assert failed unkown error\n");
							exit(0);
						}

					}
					else
					{
						fprintf(fp1, "%s --> %s = %f\n",vIdMap[pNode->Id], vIdMap[lnode->Id], vECrit[i][iFanOut]/nSimIters);
					}

					fSumECrit += ((vECrit[i][iFanOut]/nSimIters) );
					lNumWires++;
				}
			}

				fclose(fp1);
				printf(" THE TOTAL NO OF WIRES = %d  and CRITICALITY = %6.4f\n",lNumWires, fSumECrit);
			printf("The average interconnect criticality: %6.4f\n", fSumECrit / lNumWires);

			return 0;
		}
		void map_id(char* xFile)
		{
			Abc_Obj_t *lNode;
			int id =0;
			int i;
			char cmd[2048];
			FILE *fp;
			Abc_NtkForEachPi(pNtk, lNode, i)
			{
				vIdMap[id++] = strdup(Abc_ObjName(lNode));
			}
			Abc_NtkForEachPo(pNtk, lNode, i)
			{
				vIdMap[id++] = strdup(Abc_ObjName(lNode));
			}

			sprintf(cmd, "rm -rf .sens ;mkdir .sens ; grep \".names\"  %s >.sens/names.sens", xFile);
			system(cmd);
			sprintf(cmd, "awk '{c=split($0, s);  print s[c] }' .sens/names.sens > .sens/idmap.sens");

			system(cmd);
			sprintf(cmd,".sens/idmap.sens");
			fp = fopen(cmd, "r+");
			if(!fp) {
				printf("Error in reading file %s \n", cmd);
				exit(0);
			}
			sprintf(cmd,".sens/%s.n",xFile);
			FILE *fp1 = fopen(cmd, "w+");
			cmd[0] = '\0' ; 
			i=0;
			char c;

			while((c = fgetc(fp))!= EOF)
			{
				cmd[i++] = c ;
				if(c == '\n')
				{
					cmd[i-1] = '\0';
					i=0;
					vIdMap[id++] = strdup(cmd);
				}
			}
			vNumId= id;
			fclose(fp1);
			sprintf(cmd, "rm -rf .sens");
			system(cmd);
		}
		void print_mapid()
		{
			int i ;
			for(i=0; i < vNumId; i++)
				printf(" ID = %d Pin = %s \n", i, vIdMap[i]);
		}
};

int main(int argc, char*argv[])
{
	char *file_name;
	char command[1024];
	Abc_Frame_t* abc;
	int numSimVectors;

	// MANKING DETERMINITSTIC MANU srand(time(NULL));
	srand(time(NULL));

	// get the line file name
	if(argc != 3) {
		printf("Usage: sens xx.blif numVectors.\n");
		return 1;
	}

	file_name = argv[1];
	numSimVectors = atoi(argv[2]);

	// start the ABC framework
	Abc_Start();
	abc = Abc_FrameGetGlobalFrame();

	sprintf(command, "read %s", file_name);
	if(Cmd_CommandExecute(abc, command)) {
		fprintf(stdout, "Cannot execute command \"%s\".\n", command);
		return 1;
	}
	Abc_Obj_t *pNode;
	Abc_NtkSopToAig(abc->pNtkCur);


	Abc_Obj_t * fanin;
	int j;

	// simulate
	FaultSim sim(abc->pNtkCur, numSimVectors);


	//MAPPING THE NODES IN ABC TO THE REAL ONES >...
	// SHOULD USE HASH table somehow abc hash functions have problem here .. SOoo .. trying array for now .optimize later . 
		sim.map_id(file_name);
	//	sim.print_mapid();

	sim.simulate_all();

	// stop the ABC framework
	Abc_Stop();
	return 0;
}
